//=============================================================================
// Project:    SLP - An Interpreter for Super Logic Programs / Static Semantics
// Module:     ch.cxx
// Purpose:    Character classes
// Last Change:04.04.2002
// Language:   C++
// Author:     Stefan Brass
// Email:      Stefan.Brass@informatik.uni-giessen.de, sbrass@sis.pitt.edu
// Address:    Universitaet Giessen, Arndtstr. 2, 35392 Giessen, Germany
// Copyright:  (c) 1998-2001 by Stefan Brass
// Licence:    Distributed under the conditions of the GNU Public Licence.
//=============================================================================

//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"

#define	CH_MOD
#include "ch.h"

//=============================================================================
// Private Class Members:
//=============================================================================

//-----------------------------------------------------------------------------
// Initialized: This module was initialized.
//-----------------------------------------------------------------------------

	bool_t	ch_c::Initialized = BOOL_FALSE;

//-----------------------------------------------------------------------------
// Classify: Character class of each character.
//-----------------------------------------------------------------------------

	ch_t	ch_c::Classify[CH_TABSIZE];

//-----------------------------------------------------------------------------
// IsAlnum: Which characters are alphanumeric?
//-----------------------------------------------------------------------------

	bool_t	ch_c::IsAlnum[CH_TABSIZE];

//-----------------------------------------------------------------------------
// Ord: Position in "alphabet" (for sorting).
//-----------------------------------------------------------------------------

	int	ch_c::Ord[CH_TABSIZE];

//-----------------------------------------------------------------------------
// Xord: Position of second character equivalent to national character.
//-----------------------------------------------------------------------------

	int	ch_c::Xord[CH_TABSIZE];

//-----------------------------------------------------------------------------
// Xupper, Xlower: Upper and lower characters in the extended set.
//-----------------------------------------------------------------------------

#ifdef	VER_CS_MSDOS // MS-DOS Character Set:

str_t ch_c::Xupper =
"|A\216\217\265\266\267\307|B|C\200|D|E\220\322\323\324|F|G|H\
|I\326\327\330\336|J|K|L|M|N\245|O\231\235\340\342\343\345|P|Q|R|S|T\
|U\232\351\352\353|V|W|X|Y\355|Z";
str_t ch_c::Xlower =
"|a\203\204\205\206\240\306|b|c\207|d|e\202\210\211\212|f|g|h\
|i\213\214\215\241\325|j|k|l|m|n\244|o\223\224\225\233\242\344|p|q|r|s|t\
|u\201\226\227\243|v|w|x|y\230\354|z";
char ch_c::SzLig  = '\341';
char ch_c::AeUpper = '\222';
char ch_c::AeLower = '\221';

#endif

#ifdef VER_CS_ISO  // ISO 8859/1 Character Set:

str_t ch_c::Xupper =
"|A\300\301\302\303\304\305|B|C\307|D|E\310\311\312\313|F|G|H\
|I\314\315\316\317|J|K|L|M|N\321|O\322\323\324\325\326\330|P|Q|R|S|T\
|U\331\332\333\334|V|W|X|Y\335|Z";
str_t ch_c::Xlower =
"|a\340\341\342\343\344\345|b|c\347|d|e\350\351\352\353|f|g|h\
|i\354\355\356\357|j|k|l|m|n\361|o\362\363\364\365\366\370|p|q|r|s|t\
|u\371\372\373\374|v|w|x|y\375\377|z";
char ch_c::SzLig  = '\337';
char ch_c::AeUpper = '\306';
char ch_c::AeLower = '\346';

#endif

// Alternative: Standard ASCII set:

#ifndef VER_CS_ISO
#ifndef VER_CS_MSDOS
str_t ch_c::Xupper = "";
str_t ch_c::Xlower = "";
char ch_c::SzLig  = '\0';
char ch_c::AeUpper = '\0';
char ch_c::AeLower = '\0';
#endif
#endif

//=============================================================================
// Auxillary Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// ch_enter: Enter a character class into the table.
//-----------------------------------------------------------------------------

inline void ch_c::enter(char c, ch_t cl)
{
	int i = CH_CHARCAST(c);

	// Check validity of index:
	CHECK(i >= 0 && i < CH_TABSIZE, "ch_enter(index out of range)");
	CHECK(Classify[i] == CH_UNDEFINED, "ch_c::enter(class conflict)");

	// Enter character class:
	Classify[i] = cl;
}

//=============================================================================
// Exported Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

	ch_c::ch_c(void)
{
	CHECK_IMPOSSIBLE("It's not possible to create instances of ch_c");
}

//-----------------------------------------------------------------------------
// init: Initialize tables.
//-----------------------------------------------------------------------------

void	ch_c::init()
{
	register int i;
	register ch_t * p_class;
	register bool_t * p_isalnum;
	register str_t p;
	register int *p_ord;

	// Module should be initialized only once:
	CHECK(!Initialized, "ch_c::init(is already initialized)");

	// Enter default values:
	p_class = Classify;
	p_isalnum = IsAlnum;
	for(i=CH_TABSIZE; --i>=0; ) {
		*p_class++ = CH_UNDEFINED;
		*p_isalnum++ = BOOL_FALSE;
	}

	// Now enter specific values. First lowercase characters:
	for(p="abcdefghijklmnopqrstuvwxyz"; *p; p++) {
		enter(*p, CH_ATOM);
		IsAlnum[CH_CHARCAST(*p)] = BOOL_TRUE;
	}

	// Upper case characters:
	for(p="ABCDEFGHIJKLMNOPQRSTUVWXYZ_"; *p; p++) {
		enter(*p, CH_VAR);
		IsAlnum[CH_CHARCAST(*p)] = BOOL_TRUE;
	}

	// Digits:
	for(p="0123456789"; *p; p++) {
		enter(*p, CH_DIGIT);
		IsAlnum[CH_CHARCAST(*p)] = BOOL_TRUE;
	}

	// Printing characters which have no defined meaning:
	for(p="!@#^*+=\\{}[]\"`/"; *p; p++)
		enter(*p, CH_UNUSED);

	// Special characters:
	enter('\0', CH_NULL);
	enter('\004', CH_EOF); // Crtl+D
	enter('\032', CH_EOF); // Crtl+Z
	enter('-', CH_MINUS);
	enter('\'', CH_QUOTE);
	enter('.', CH_FULLSTOP);
	enter(',', CH_COMMA);
	enter(':', CH_COLON);
	enter(';', CH_SEMICOLON);
	enter('(', CH_LPAREN);
	enter(')', CH_RPAREN);
	enter('~', CH_NOT);
	enter('&', CH_AND);
	enter('|', CH_OR);
	enter('<', CH_LESS);
	enter('>', CH_GREATER);
	enter('?', CH_QUEST);
	enter('$', CH_CMD);

	// Lowercase national characters:
	CHECK_CODE(i = 0);
	for(p=Xlower; *p; p++) {
		if(*p=='|') {
			p++; // the loop increment skips the ASCII letter
			CHECK(i<26 && *p=="abcdefghijklmnopqrstuvwxyz"[i++],
				"ch_c::init(bad ch_c::Xlower)");
		}
		else {
			enter(*p, CH_ATOM);
			IsAlnum[CH_CHARCAST(*p)] = BOOL_TRUE;
		}
	}
	// ss and ae see below

	// Uppercase national characters:
	CHECK_CODE(i = 0);
	for(p=Xupper; *p; p++) {
		if(*p=='|') {
			p++; // the loop increment skips the ASCII letter
			CHECK(i<26 && *p=="ABCDEFGHIJKLMNOPQRSTUVWXYZ"[i++],
				"ch_c::init(bad ch_c::Xlower)");
		}
		else {
			enter(*p, CH_VAR);
			IsAlnum[CH_CHARCAST(*p)] = BOOL_TRUE;
		}
	}
	// Ae see below

	// Now the alphabetic positions for sorting:
	p_ord = Ord;
	for(i=0; i<CH_TABSIZE; i++)
		*p_ord++ = i;

	// Make the code of the NUL character 0:
	Ord[CH_CHARCAST('\0')] = 0;
		// This is currently redundant.
		// But the old version of the CH_CHARCAST macro added 128.

	i = 256;	// Even on Non-ASCII systems, we use ASCII order
	for(p=" !\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~"; *p; p++)
		Ord[CH_CHARCAST(*p)] = i++;

	i = 512;
	for(p="0123456789"; *p; p++)
		Ord[CH_CHARCAST(*p)] = i++;

	i = 1001;
	for(p="abcdefghijklmnopqrstuvwxyz"; *p; p++)
		Ord[CH_CHARCAST(*p)] = i++;

	i = 1001;
	for(p="ABCDEFGHIJKLMNOPQRSTUVWXYZ"; *p; p++)
		Ord[CH_CHARCAST(*p)] = i++;

	i = 1000;
	for(p=Xlower; *p; p++) {
		if(*p=='|') {
			p++; // the loop increment skips the ASCII letter
			i++;
		}
		else
			Ord[CH_CHARCAST(*p)] = i;
	}

	i = 1000;
	for(p=Xupper; *p; p++) {
		if(*p=='|') {
			p++; // the loop increment skips the ASCII letter
			i++;
		}
		else
			Ord[CH_CHARCAST(*p)] = i;
	}

	p_ord = Xord;
	for(i=0; i<CH_TABSIZE; i++)
		*p_ord++ = 0;

	// Finally special "double" characters:
	if(SzLig!=0) {
		enter(SzLig, CH_ATOM);
		IsAlnum[CH_CHARCAST(SzLig)] = BOOL_TRUE;
		Ord [CH_CHARCAST(SzLig)] = 1019;
		Xord[CH_CHARCAST(SzLig)] = 1019;
	}
	if(AeLower!=0) {
		enter(AeLower, CH_ATOM);
		IsAlnum[CH_CHARCAST(AeLower)] = BOOL_TRUE;
		Ord [CH_CHARCAST(AeLower)] = 1001;
		Xord[CH_CHARCAST(AeLower)] = 1005;
	}
	if(AeUpper!=0) {
		enter(AeUpper, CH_ATOM);
		IsAlnum[CH_CHARCAST(AeUpper)] = BOOL_TRUE;
		Ord [CH_CHARCAST(AeUpper)] = 1001;
		Xord[CH_CHARCAST(AeUpper)] = 1005;
	}

	// Ok, module initialized:
	Initialized = BOOL_TRUE;
}

